classdef scenario < handle
    %Scenario builds a room and generates random walks
    %   Represents a virtual scenario where user walks
    
    properties
        area    % (#)  dimension
        len     % (#)  length of scenario
        wid     % (#)  width of scenario
        dep     % (#)  depth
        scene   % ([]) virtual representation
        unit    % (#)  quantaty to identify measure in Matlab
        ndim
        objs    % ([]) objects location and type
        Mx
        My
        Mz
        coord  
        limits  % ([]) limits of the area
        exists  % (b) flag that mentions if it exists or not
        
        dir     % storage directory
        storage % matfile to store data
    end
    
    
    
    
    
    methods(Static)
        
        function pos = coordinateMapping(pos,region_limits)

            pos = fix(pos+1);
            pos(pos(:,1) < region_limits(1,1)+1) = region_limits(1,1)+1;% zero does not count
            pos(pos(:,1) > region_limits(2,1)+1) = region_limits(2,1)+1;% zero does not count
            
            pos(pos(:,2) < region_limits(1,2)+1) = region_limits(1,2)+1;% zero does not count
            pos(pos(:,2) > region_limits(2,2)+1) = region_limits(2,2)+1;% zero does not count
            
        end
        
        function nextStep = movementsAvailable(ndim)
            
            choices = [ -1, 0 , 0;
                1 , 0 , 0;
                0 , 1 , 0;
                0 , -1, 0;
                0 , 0 , 1;
                0 , 0 , -1];
            
            n = ndim*2;
            if n < size(choices,1)
                nextStep = choices(1:n,1:ndim);
            else
                error('Dimension not supported');
            end
        end
        
        
        function walk = ramdomWalk(nsteps,limits,steplength)            
            % creates starting point
            walk = nan(nsteps,2);
            dim  = size(limits,2);
            
            % check if 3D that it is not 0
            if limits(1,end)-limits(2,end) == 0 && dim == 3
                dim = 2;
            end
            
            % initial point
            walk(1,:) = math.drawRandomNdimNumber( limits );
            
            % Retrieves available motion and scales it for step length
            moves = build.scenario.movementsAvailable(dim);
            moves = moves.*steplength;
            
            % generate choices vector
            % basically a vector to index choices
            choices = moves(math.drawDiscrete(dim*2,nsteps-1,1),:);
            for k=1:nsteps-1 % creates path
                walk(k+1,:)=walk(k,:)+choices(k,:);
                cnt = 0;
                %validate step
                while ~geometry.isinside(walk(k+1,:),limits) %redraw
                    walk(k+1,:)=walk(k,:)+moves(math.drawDiscrete(dim*2,1,1),:)*steplength;
                    cnt = cnt + 1;
                    if cnt > 100
                        error('build:scenario','stuck inside random walk');
                    end
                end
                assert(walk(k+1,1) == walk(k,1) || walk(k+1,2) == walk(k,2));
            end
            
        end
        
        function [Mx,My,Mz,limits] = buildCoordinates(len,wid,dep,matlab_unit)
            % BUILDCOORDINATES provides a coordinate system for the given
            % area in the matlab unit system provided (converted from
            % meter)
            %
            %   LEN is defined as left to right (# cols)
            %   WID is defined as bottom to top (# rows)
            %
            %   ORIG defines the LOCATION of the origin, either top left or
            %   bottom left (No negative numbering here!)
            %
            %   Example,
            %       If the grid size is 0.5m then matlab_unit should be 2
            %
            
            % wid defines the number of rows, therefore the values of x
            % len defines the number of colm, therefore the values of y

            My = repmat((0:matlab_unit:len)',1,len)';
            Mx = repmat((0:matlab_unit:wid)',1,wid);
            Mz = repmat((0:matlab_unit:dep)',1,dep);
            limits = [0,0,0;wid,len,dep];
            
            if isempty(Mz)
                Mz = zeros(size(Mx));
            end
            
        end
        
        
        function [wid,len,dep] = computeWidLen(limits)
           wid = limits(2,1)-limits(1,1);%x
           len = limits(2,2)-limits(1,2);%y 
           dep = limits(2,3)-limits(1,3);%z  
        end
        
    end
    
    
    methods(Access = private)
       function pos = place_object(obj,limits,restricted_area)
            
            % Creates aps in the grid
            pos = math.drawRandomNdimNumber( limits(:,1:obj.ndim) );
            
            % check if valid for each room!
            if ~isempty(restricted_area)
                redraw = 1;
                while redraw
                    if(geometry.isinside(pos,restricted_area))
                        pos = math.drawRandomNdimNumber( limits(:,1:obj.ndim) );
                    else
                        redraw = 0;
                    end
                end
            end
        end
         
    end
    
    methods
        function obj = scenario(len,wid,dep,ndim,unit,dir,matfilename,forceOW)
            % SCENARIO is the object constructor
            %   AREA and UNIT are inputs in meter
            obj.len   = len;
            obj.wid   = wid;
            obj.dep   = dep; if isempty(dep), obj.dep = 0; end;
            obj.ndim  = ndim;
            obj.area  = len*wid;
            obj.unit  = 1/unit; % convert meter to vector size
            obj.scene = zeros(wid+1,len+1); % floor map
            obj.dir   = dir;
            
            if nargin >5
                % initialise storage
                if exist([dir,matfilename,'.mat'],'file') && ~forceOW
                    obj.exists  = true;
                    obj.storage = inout.output.createMatfile(dir,matfilename,true,false); % Writable and Overwrite
                else
                    obj.exists  = false;
                    obj.storage = inout.output.createMatfile(dir,matfilename,true,true); % Writable and Overwrite
                end
            end
            % Creates a coordinate system
            obj.createCoordinates();
            
            obj.storage.coordinates = {obj.Mx,obj.My,obj.Mz,obj.limits};            
        end
       
        
        function setReference(obj,main_limits)   
            n_wid = main_limits(2,1)-main_limits(1,1);%x
            n_len = main_limits(2,2)-main_limits(1,2);%y 
            obj.scene = zeros(n_wid+1,n_len+1); % floor map
            obj.coord = main_limits;
        end
        
        function mapToScene(obj,pos_vec,region_limits,type)
            
            pos_vec = build.scenario.coordinateMapping(pos_vec,region_limits);
            for k=1:size(pos_vec,1)
                obj.scene(pos_vec(k,1),pos_vec(k,2))=type(k); 
            end
            
        end
        
        
        function fixObjects(obj,type,location)
            % fixes object to given coordinates
            region = obj.limits; %assumes objects limits
            posN = location;
            obj.mapToScene(posN,region,type);
            obj.objs = [posN,type];
            obj.storage.objs   = obj.objs;
            obj.storage.nbObjs = size(obj.objs,1);
        end
        
        function posN = placeObjects(obj,N,type,invalid_limits)
            region = obj.limits; %assumes objects limits
            posN = NaN(N,obj.ndim);
            for k=1:N
                posN(k,:) = obj.place_object(region,invalid_limits);
            end
            obj.mapToScene(posN,region,type);
            obj.objs           = [posN,type];
            obj.storage.objs   = obj.objs;
            obj.storage.nbObjs = size(obj.objs,1);
            obj.storage.scene  = obj.scene;
        end
        
        function buildScene(obj,area)
            obj.dim = area;
        end
        
        function createCoordinates(obj)
            [obj.Mx,obj.My,obj.Mz,obj.limits] = build.scenario.buildCoordinates(obj.len,obj.wid,obj.dep,obj.unit);
        end
        
        function create2DWalk(obj,nbSteps,lenSteps,nbWalks,maxEm)
            %CREATE2DWALK generates N walks of size nbSteps with a step
            %size defined by lenSteps.
            % The walks are stored in a matfile.
            
            obj.storage.walk(nbSteps,2,nbWalks)=NaN;
            obj.storage.number_of_walks = nbWalks;
            obj.storage.number_of_steps_per_walk = nbSteps;
            obj.storage.ermse(nbSteps,6,nbWalks,maxEm,maxEm,maxEm,maxEm,4)=NaN; %this is to specificy for a general implementation
            obj.storage.estimates(nbSteps,obj.ndim,6,nbWalks,maxEm,maxEm,maxEm,maxEm,4)=NaN; %this is to specificy for a general implementation
            for k=1:nbWalks
                obj.storage.walk(:,:,k) = build.scenario.ramdomWalk(nbSteps,obj.limits(:,1:2),lenSteps);
            end
        end
        
        %TODO: make it more generic
        function ranging(obj,Freq,BW,G,Nfig,F2,model,To)
            S = obj.storage.number_of_steps_per_walk;
            P = obj.storage.number_of_walks;
            N = obj.storage.nbObjs;
            
            obj.storage.ranges(N,S,P) = NaN;
            obj.storage.cno(N,S,P)    = NaN;
            obj.storage.svar(N,S,P)   = NaN;
            
            ranges(N,S)               = NaN;
            cno(N,S)                  = NaN;
            svar(N,S)                 = NaN;
            measurements(N,S)         = NaN;
            
            % for each path
            for p = 1:P
                walk = obj.storage.walk(:,:,p);
                for s = 1:S
                    ranges(:,s)   = math.distance(walk(s,:),obj.objs(:,1:2));
                    for n=1:N %Computes CN0 and retrieves variance for signal
                        [cno(n,s),svar(n,s)] = RadioSpectrum.Pathloss.dist2cno(ranges(n,s),Freq(n),BW(n),G(n),Nfig(n),F2(n),model(n,:),To);
                        measurements(n,s)    = normrnd(0,sqrt(svar(:,s)));% Creates measurements
                    end
                end
                % vectorises storage
                % obtain correct matrix format by reshaping to
                % ranges = reshape(ranges,N,S,P)
                % cno    = reshape(cno,N,S,P)
                % svar   = reshape(svar,N,S,P)
                obj.storage.ranges(:) = ranges(:);
                obj.storage.cno(:)    = cno(:);
                obj.storage.svar(:)   = svar(:);
            end
        end
        
    end
    
end