classdef spectrum < handle
    %SPECTRUM<HANDLE aggregates spectrum information
    %
    %   CONSTRUCTOR
    %    Spectrum()
    %
    %
    %   Tampere University of Technology
    %   Pedro Silva, 2014
    %
    %   TODO
    %   Currently RSS can have different size! make sure it is the
    %   reference size
    
    %%% VARIABLES
    %%%%% PRIVATE
    %%%%% PUBLIC
    %%%%% PROTECTED
    properties (Constant)
        wifi  = 1;
        rfid  = 2;
        blue  = 3;
        uwb   = 4;
        light = 5;
        lte   = 6;
    end
    
    properties (Constant, Access = private)
        idx_train = 1;  % (#) training index for shadowing matrix
        idx_acq   = 2;  % (#) acquisition index for shadowing matrix
    end
    
    properties(Access = protected)
        id_counter % ([]) next free id
        radiomaps  % ([]) matrix containing all the RADIO     maps for each emitter/technology
        shadowing  % ([]) matrix containing all the N-SHADOWING maps for each emitter/technology
        limits     % ([]) contains the limits of the area
        unit       % (#)  how many meters each position should represent
        emitter  % ([]) contains list of ID, techtype and positon?
        rm_size    % (#)  radio maps size
        rm_pos     % ([]) points acquired in the database
    end
    
    properties(Dependent)
        nbEmitters % (#) Total number of emitter being considered
    end
    
    properties
        x           % x coordinates
        y           % y coordinates
        z           % z coordinates
    end
    
    
    %%%%% ABSTRACT
    methods(Abstract)
        survey(obj,varargin);
        %         acquire(obj,track);
    end
    
    %%% CONSTRUCTOR
    methods
        function obj = spectrum(limits,unit,varargin)
            obj.limits     = limits;
            obj.unit       = unit;
            obj.id_counter = 1;
            
            % Compute starting and ending points
            scale       = obj.unit;
            limit       = obj.limits;
            x1          = limit(1,1); %x1
            y1          = limit(1,2); %x2
            z1          = limit(1,3); %x2
            x2          = limit(2,1); %y1
            y2          = limit(2,2); %y2
            z2          = limit(2,3); %x2
            
            % Generates a matrix table with distance
            xlen = length(x1:scale:x2);
            ylen = length(y1:scale:y2);
            zlen = length(z1:scale:z2);
            
            obj.x = repmat(x1:scale:x2,xlen,1)';
            obj.y = repmat(y1:scale:y2,ylen,1);
            if zlen ~= xlen
                obj.z = obj.x.*0;
                obj.rm_size = [xlen ylen];
            else
                obj.z = repmat(z1:scale:z2,zlen,1);
                obj.rm_size = [xlen ylen zlen];
            end
        end
        function reset(obj)
            obj.radiomaps=[];
            obj.shadowing=[];
        end
    end
    
    %%% METHODS
    %%%%% PRIVATE
    methods(Access = private, Static)
        
    end
    
    methods(Access=private)
        function wipeRadioMap(obj,id,orientation)
            Em  = obj.emitter;
            pos = Em{2}(id,1:3);
            x = fix(pos(1));
            y = fix(pos(2)); 
%             totrim = reshape(obj.radiomaps(:,id),max(obj.rm_size(1)),max(obj.rm_size(2)));
            
            switch(orientation)
                case 'up'
                    obj.radiomaps(obj.y < pos(2),id) = NaN;
                case 'down'
                    obj.radiomaps(obj.y > pos(2),id) = NaN;
                case 'left'
                    obj.radiomaps(obj.x > pos(1),id) = NaN;
                case 'right'
                    obj.radiomaps(obj.x < pos(1),id) = NaN;
                case '90-right-down'
                    obj.wipeRadioMap('right');
                    obj.wipeRadioMap('down');
                case '90-left-down'
                    obj.wipeRadioMap('left');
                    obj.wipeRadioMap('down');
                case '90-right-up'
                    obj.wipeRadioMap('right');
                    obj.wipeRadioMap('up');
                case '90-left-up'
                    obj.wipeRadioMap('left');
                    obj.wipeRadioMap('up');
            end
%             obj.radiomaps(:,id) = totrim(:);
        end
    end
    
    %%%%% PUBLIC
    methods(Static)
        
        
        function str = getInterfaceName(intId)
            switch(intId)
                case radio.spectrum.wifi
                    str = 'WIFI';
                case radio.spectrum.rfid
                    str = 'RFID';
                otherwise
            end
        end
        
        function [IntId,hFunc] = getInterfaceIdentity(strList)
            %GETINTERFACEID converts text to integer represenation
            %
            %   How to access the handles: hFunc(k) <= hFunc{3}{1}(k)
            %
            lenList          = size(strList,1);
            IntId(lenList,1) = -1;
            hFunc            = cell(size(strList,1),1);
            for k=1:lenList
                switch(upper(strList(k,1:3)))
                    case 'WIF', %WIFI
                        IntId(k) = radio.spectrum.wifi;
                        hFunc{k} = [{@radio.pathloss.wifi_model_params};
                            {@radio.pathloss.wifi_rss};...
                            {@radio.pathloss.wifi_shadowing}];
                    case 'RFI', %RFID
                        IntId(k)   = radio.spectrum.rfid;
                        hFunc{k}   = [{@radio.pathloss.rfid_model_params};
                            {@radio.pathloss.rfid_rss};...
                            {@radio.pathloss.rfid_shadowing}];
                    case 'BLU', %BLUETOOTH
                        IntId(k) = radio.spectrum.blue;
                        hFunc{k} = NaN;
                    case 'UWB', %UWB
                        IntId(k) = radio.spectrum.uwb;
                        hFunc{k} = NaN;
                    case 'LTE', %LTE
                        IntId(k) = radio.spectrum.lte;
                        hFunc{k} = NaN;
                    otherwise
                        error('spectrum:setInterfaceIdentity','radio type unknown');
                end
            end
            hFunc = {IntId, hFunc}; %tag function with interface id
        end
        
        function IntId = inter2id(strList)
            %GETINTERFACEID converts text to integer represenation
            lenList = size(strList,1);
            IntId(lenList,1) = -1;
            for k=1:lenList
                switch(upper(strList(k,1:3)))
                    case 'WIF', %WIFI
                        IntId(k) = 1;
                    case 'RFI', %RFID
                        IntId(k) = 2;
                    case 'BLU', %BLUETOOTH
                        IntId(k) = 3;
                    case 'UWB', %UWB
                        IntId(k) = 4;
                    case 'LTE', %LTE
                        IntId(k) = 5;
                    otherwise
                        error('radio type unknown');
                end
            end
        end
    end
    
    
    
    
    %%%%% PUBLIC
    methods
        
          function fixDirection(obj,id,location)
            
            nbEms = length(id);
            
            if strcmpi(location,'CORNER')
                %do nothing
                return
            end
            
            switch (nbEms)
                
                case 1
                    obj.direction(id,'up');
                    
                case 2
                    obj.direction(id(1),'up');
                    obj.direction(id(2),'down');
                    % correct
                    
                case 4
                    % i do not think it is necessary to switch according to middle
                    obj.direction(id(1),'up');
                    obj.direction(id(2),'down');
                    obj.direction(id(3),'right');
                    obj.direction(id(4),'left');
            end
            
        end
        
        
        function direction(obj,id,direction)
            obj.wipeRadioMap(id,direction);
        end
        
        function create(obj,type,emitter)
            %CREATE creaes N radio maps for each emmitter given
            % [(x,y,(z)),tech,id]
            bEm = NaN(size(emitter,1),size(emitter,2)+2);
            
            % Update the counter
            for id = obj.id_counter:length(emitter)
                
                bEm(id,:) = [emitter(id,:),id];
                
                % create radio map
                obj.survey(emitter(1:3,type(id,:)));
                
            end
            
            % Store information
            obj.id_counter = id+1; % creates next ID
            if isempty(obj.emitter)
                obj.emitter = bEm;
            else
                obj.emitter = [obj.emitter,bEm];
            end
            
        end
        
        
        function insert(obj,id,radio_map,shadowing_map)
            %INSERT stores in the database the radio maps
            %   MAPS are vectorised!
            obj.radiomaps(:,id) = radio_map;
            obj.shadowing(:,id) = shadowing_map;
        end
        
        %         function meas = probe(obj,pos,id,varargin)
        %             %PROBE rerieves the value at the given position
        %              idx  = sub2ind(obj.rm_size, pos(1), pos(2),pos(3));
        %              meas = obj.radiomaps(idx,id);
        %         end
        
        
        function map = retrieve(obj,id,ideal)
            %RETRIEVE returns the maps with corrsponding shadowing
            %  Each radiomap is a collumn
            %  There is 2 collumns for the shadowing
            map = obj.radiomaps(:,id);
            if ~ideal % if true, add no shadowing
                map = map + obj.shadowing(:,id);
            end
        end
        
        
        % Make a copy of a handle object.
        function new = copy(this)
            % Instantiate new object of the same class.
            new = feval(class(this));
            
            % Copy all non-hidden properties.
            p = fieldnames(struct(this));;
            for i = 1:length(p)
                new.(p{i}) = this.(p{i});
            end
        end
        
        function heard = getHeardAps(obj,pos)
            idx   = sub2ind(obj.rm_size,pos(1),pos(2));
            heard = obj.radiomap(idx,:)+obj.shadowing(idx,:);
        end
        
        function [heard,nbSeen] = dumpSelection(obj,sel_ids,sensitivity)
            nbSeen = length(sel_ids);
            heard  = obj.radiomaps(:,sel_ids)'+obj.shadowing(:,sel_ids)';
            heard  = heard(:);
            if nargin == 3 && ~isempty(sensitivity)
                sensitivity = repmat(sensitivity,length(obj.rm_pos),1);
                heard(heard < sensitivity) = NaN;
            end
        end
        
        function heard = dumpHeardAps(obj,sensitivity)
            heard = obj.radiomaps'+obj.shadowing';
            % Saturates to sensitivity
            if nargin == 2 && ~isempty(sensitivity)
                for k=1:size(heard,2) % for each collumn
                    heard(heard(:,k) < sensitivity,k) = NaN;
                end
            end
            heard = heard(:); % vector with all APs measurement for each point in the data base            
        end
        
    end
    %%%%% PROTECTED
    
    
    %%% GETTERS and SETTERS
    methods
        function em = getDB(obj)
            em = obj.emitter;
        end
        
        function nb = get.nbEmitters(obj)
            nb = length(obj.emitter{1});
        end
        
        function pos = getCoordinates(obj)
            pos = obj.rm_pos;
        end
        
    end
    
    methods(Access = protected)
        function setDB(obj,em)
            obj.emitter = em;
        end
    end
    
    
end

